#
# Copyright (c) SAS Institute Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import os
import posixpath

from six.moves.urllib.parse import urlparse
import requests
import six

from . import utils
from .artifact import Artifact
from .errors import MissingArtifactError
from .errors import MissingPathError
from .pom import Pom
from .versioning import VersionRange

try:
    from xml.etree import cElementTree as ElementTree
except ImportError:
    from xml.etree import ElementTree

log = logging.getLogger(__name__)


class MavenClient(object):
    """ Client for talking to a maven repository
    """
    def __init__(self, *urls):
        if isinstance(urls, six.string_types):
            urls = [urls]
        self._repos = []
        for url in urls:
            url = urlparse(url)
            if not url.scheme or url.scheme == "file":
                self._repos.append(LocalRepository(url.path))
            elif url.scheme.startswith("http"):
                self._repos.append(HttpRepository(url.geturl()))
            else:
                msg = "Unknown scheme: %s"
                log.error(msg, url)
                raise ValueError(msg % url.geturl())

    def find_artifacts(self, coordinate):
        """Find all artifacts matching the coordinate

        :param str coordinate: maven coordinate
        :retrun: list of Artifacts
        """
        artifacts = set([])
        for repo in self._repos:
            version = repo.get_versions(coordinate)
            if version is not None:
                artifacts.update(set(version))
        return sorted(artifacts, reverse=True)

    def get_metadata(self, coordinate):
        """Return the metadata associated with the coordinates

        :param str coordinate: maven coordinate
        :raises: :py:exc:`pymaven.errors.MissingArtifactError`
        :return: the metadata requested
        :rtype: :py:class:`pymaven.pom.Pom`
        """
        query = Artifact(coordinate)
        assert query.version.version is not None, \
            "Cannot get metadata for version range"

        if query.type != "pom":
            query.type = "pom"

        for repo in self._repos:
            if repo.exists(query.path):
                return Pom(coordinate, self)
        else:
            raise MissingArtifactError(coordinate)

    def get_artifact(self, coordinate):
        """Return the actual artifact specified by the coordinate

        :param str coordinate: maven coordinate
        :raises: :py:exc:`pymaven.errors.MissingArtifactError`
        :return: the artifact requested
        :rtype: :py:class:`pymaven.Artifact`
        """
        query = Artifact(coordinate)
        assert query.version.version is not None, \
            "Cannot get artifact for version range"

        for repo in self._repos:
            if repo.exists(query.path):
                query.contents = repo.open(query.path)
                return query
        else:
            raise MissingArtifactError(coordinate)


class AbstractRepository(object):
    """
    Abstract class, this defines the interface that all repositories implement
    """
    def __init__(self, url):
        self._url = url

    def get_versions(self, coordinate):
        query = Artifact(coordinate)
        if self.exists(query.path):
            if query.version and query.version.version:
                return [Artifact(coordinate)]

            version_range = query.version
            if version_range is None:
                version_range = VersionRange("[,)")

            # base coordinate for all return values is everything up to the
            # version of the query
            base_coordinate = "%s:%s" % (query.group_id, query.artifact_id)
            if query.classifier:
                base_coordinate += "%s:%s" % (query.type, query.classifier)
            elif query.type != "jar":
                base_coordinate += ":%s" % query.type

            return sorted([
                Artifact(':'.join([base_coordinate, version]))
                for version in self.listdir(query.path)
                if version in version_range], reverse=True)

    def exists(self, path):
        """Return ``True`` if *path* exists in the repository, ``False``
        otherwise
        """
        return self._exists(path)

    def listdir(self, path):
        """List contents of *path*

        :raises: MissingPathError if *path* does not exist
        """
        try:
            return self._listdir(path)
        except (OSError, requests.exceptions.HTTPError):
            raise MissingPathError(path)

    def open(self, path):
        try:
            return self._open(path)
        except (IOError, requests.exceptions.HTTPError):
            raise MissingPathError("No such file: %s" % path)


class HttpRepository(AbstractRepository):
    """ Access a maven repository via http
    """
    def __init__(self, url, username=None, password=None):
        super(HttpRepository, self).__init__(url)

    def _get(self, uri, **kwargs):
        res = self._request("GET", uri, **kwargs)
        return res

    def _head(self, uri, **kwargs):
        res = self._request("HEAD", uri, **kwargs)
        return res

    def _request(self, method, uri, json=False, **kwargs):
        url = utils.urljoin(self._url, uri)
        log.debug("requesting %s %s", method, url)
        res = requests.request(method, url, **kwargs)

        if res.status_code != requests.codes.ok:
            raise requests.HTTPError(res.reason)

        if json:
            return res.json()
        return res

    def _exists(self, path):
        try:
            self._head(path)
        except requests.exceptions.HTTPError:
            return False
        return True

    def _listdir(self, path):
        uri = posixpath.join(path, "maven-metadata.xml")
        res = self._get(uri)
        with res as fh:
            metadata = ElementTree.parse(fh)

        return [elem.text.strip()
                for elem in metadata.findall("versioning/versions/version")
                if elem is not None]

    def _open(self, path):
        res = self._get(path, stream=True)
        return res


class LocalRepository(AbstractRepository):
    """A local disk-based repository
    """
    def _join(self, *args):
        return os.path.join(self._url, *args)

    def _exists(self, path):
        return os.path.exists(self._join(path))

    def _listdir(self, path):
        return os.listdir(self._join(path))

    def _open(self, path):
        """Open *path* as a file-like object

        The caller is responsible for calling close on the object
        """
        return open(self._join(path))
